"""aws-s3."""

from pathlib import Path

import rich_click as click

from aws_s3 import __version__

click.rich_click.USE_MARKDOWN = True
OUTPUT_FILE_DEFAULT = "output"


@click.command()
@click.version_option(version=__version__, prog_name="aws-s3")
@click.argument("input_file", type=click.Path(exists=True))
@click.argument("output_file", type=click.Path(), required=False)
@click.option("--force", is_flag=True, help="Overwrite the output file if it exists.")
def aws_s3(input_file: str, output_file: str, force: bool) -> None:
    """
    `INPUT_FILE` to `OUTPUT_FILE`.
    """
    output_path = (
        Path(output_file)
        if output_file
        else Path(input_file).parent / f"{OUTPUT_FILE_DEFAULT}.txt"
    )
    if output_path.exists() and not force:
        click.echo(f"Output file {output_path} already exists. Use --force to overwrite.")
        raise click.Abort()

    click.echo(f"File saved to {output_path}")


if __name__ == "__main__":  # pragma: no cover
    aws_s3()  # pylint: disable=no-value-for-parameter



import os
import pathlib
import socket
import subprocess
import time
from unittest.mock import patch

import pytest
import yaml

import boto3
import aiobotocore
import aiobotocore.session


@pytest.fixture(scope="session", autouse=True)
def set_fake_aws_credentials():
    """Set fake AWS credentials for the test session."""
    os.environ['AWS_ACCESS_KEY_ID'] = 'fake_access_key'
    os.environ['AWS_SECRET_ACCESS_KEY'] = 'fake_secret_key'
    os.environ['AWS_SESSION_TOKEN'] = 'fake_session_token'
    yield


def create_s3_folder(bucket_name, structure, s3_client, parent_path=""):
    if isinstance(structure, list):
        for item in structure:
            for key, value in item.items():
                create_s3_objects(key, value, bucket_name, s3_client, parent_path)
    else:
        for key, value in structure.items():
            create_s3_objects(key, value, bucket_name, s3_client, parent_path)


def create_s3_objects(key, value, bucket_name, s3_client, parent_path):
    if isinstance(value, list):
        for item in value:
            if isinstance(item, str) and item == "EMPTY":
                folder_path = os.path.join(parent_path, key)
                s3_client.put_object(Bucket=bucket_name, Key=(folder_path + "/"))
            elif isinstance(item, str):
                file_path = os.path.join(parent_path, key, item)
                s3_client.put_object(Bucket=bucket_name, Key=file_path, Body="")
            elif isinstance(item, dict):
                folder_path = os.path.join(parent_path, key)
                create_s3_folder(bucket_name, item, s3_client, folder_path)
    elif isinstance(value, dict):
        folder_path = os.path.join(parent_path, key)
        create_s3_folder(bucket_name, value, s3_client, folder_path)


def wait_for_moto_server(s3_client, moto_server_process, retries=5, delay=1):
    for i in range(retries):
        try:
            s3_client.list_buckets()
            return True
        except Exception as exc:
            print(f"Attempt {i + 1}/{retries} failed: {exc}")
            stderr_output = moto_server_process.stderr.readline()
            if stderr_output:
                print(stderr_output.decode(), end="")
            time.sleep(delay)
    return False


def get_free_port():
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(('', 0))
        return s.getsockname()[1]


@pytest.fixture(scope="session")
def fake_s3_server():
    """Starts a Moto server.

    Moto does not support mocking async clients, so we have to use fake server.

    Returns (<S3 client>, <factory for async S3 clients>).
    """
    port = get_free_port()
    moto_server_process = subprocess.Popen(
        ["moto_server", "-H", "0.0.0.0", f"-p{port}"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    endpoint_url = f"http://127.0.0.1:{port}"
    s3_client = boto3.client("s3", region_name="us-east-1", endpoint_url=endpoint_url)

    if not wait_for_moto_server(s3_client, moto_server_process):
        print("Failed to start the Moto server.")
        moto_server_process.terminate()
        moto_server_process.wait()
        pytest.fail("Failed to start Moto server")

    session = aiobotocore.session.get_session()
    yield (
        s3_client,
        lambda: session.create_client(
            "s3", region_name="us-east-1", endpoint_url=endpoint_url
        ),
    )

    # Teardown
    moto_server_process.terminate()
    moto_server_process.wait()


@pytest.fixture(scope="session")
def mock_s3_structure(fake_s3_server):
    s3_client, s3_async_client_factory = fake_s3_server
    HERE = pathlib.Path(__file__).parent
    with (HERE / "resources" / "bucket_keys.yml").open("r") as file:
        structure = yaml.safe_load(file)
    bucket_name = "mock-bucket"
    s3_client.create_bucket(Bucket=bucket_name)
    create_s3_folder(bucket_name, structure, s3_client)
    with patch("aws_s3.list_objects_async.get_s3_client") as mock_client:
        mock_client.side_effect = s3_async_client_factory
        yield



import asyncio
from aws_s3 import ListObjectsAsync


async def main():
    bucket_name = "your-bucket-name"
    prefix = "your-prefix/"
    list_objects = ListObjectsAsync(bucket_name)
    objects = await list_objects.list_objects(prefix=prefix, max_depth=2)

    for obj in objects:
        print(obj["Key"])

if __name__ == "__main__":
    asyncio.run(main())



"""S3 Bucket helper utils. Async list objects by folders

The file is mandatory for build system to find the package.
"""

from aws_s3.__about__ import __version__
from aws_s3.list_objects_async import ListObjectsAsync

__all__ = ["__version__", "ListObjectsAsync"]



#!/usr/bin/env python3
"""Include requirements.txt in pyproject.toml project.dependencies array.

Author: Andrey Sorokin
email: andrey@sorokin.engineer
url: https://github.com/andgineer
License: MIT
"""

import argparse

import toml

PROJECT_METADATA_FILE_NAME = "pyproject.toml"
REQUIREMENTS_FILE_NAME = "requirements.txt"


def main(requirements_file_name: str, section_path: str) -> None:
    """Include requirements.txt in pyproject.toml project.dependencies array."""
    with open(requirements_file_name, "r", encoding="utf8") as f:
        requirements = [
            line.strip() for line in f if line.strip() and not line.lstrip().startswith("#")
        ]
    print(
        f"From {requirements_file_name} read requirements:"
        f"\n{requirements[:10]}...\n...{len(requirements)} total"
    )

    with open(PROJECT_METADATA_FILE_NAME, "r", encoding="utf8") as f:
        pyproject_data = toml.load(f)

    section_keys = section_path.split(".")
    target_section = pyproject_data
    for key in section_keys:
        target_section = target_section[key]

    print(
        f"To section `{section_path}` array `dependencies`"
        f" of `{PROJECT_METADATA_FILE_NAME}`, project `{pyproject_data['project']['name']}`.\n"
    )
    target_section["dependencies"] = requirements

    with open(PROJECT_METADATA_FILE_NAME, "w", encoding="utf8") as f:
        toml.dump(pyproject_data, f)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description="""Include dependencies from `requirements_file` into array `dependencies`
of pyproject.toml's section specified in `section_path`.

By default includes requirements.txt into `dependencies` array of section `project`.""",
    )
    parser.add_argument(
        "requirements_file",
        nargs="?",
        default=REQUIREMENTS_FILE_NAME,
        help="Input file with dependencies",
    )
    parser.add_argument(
        "section_path",
        nargs="?",
        default="project",
        help="Path to the section in pyproject.toml (e.g., tool.hatch.envs.test)",
    )

    args = parser.parse_args()
    main(requirements_file_name=args.requirements_file, section_path=args.section_path)



import os
import shutil
import sys

from invoke import task, Context, Collection
import subprocess


def get_allowed_doc_languages():
    build_docs_file_name = "scripts/build-docs.sh"
    try:
        with open(build_docs_file_name, "r") as f:
            for line in f:
                if "for lang in" in line:
                    langs = line.split("in")[1].strip().split(";")[0].split()
                    return [lang.strip() for lang in langs]
    except FileNotFoundError:
        print(f"No {build_docs_file_name} file found")
    return ["en", "bg", "de", "es", "fr", "ru"]  # default


ALLOWED_DOC_LANGUAGES = get_allowed_doc_languages()
ALLOWED_VERSION_TYPES = ["release", "bug", "feature"]


@task
def version(c: Context):
    """Show the current version."""
    with open("src/aws-s3/__about__.py", "r") as f:
        version_line = f.readline()
        version_num = version_line.split('"')[1]
        print(version_num)
        return version_num


def ver_task_factory(version_type: str):
    @task
    def ver(c: Context):
        """Bump the version."""
        c.run(f"./scripts/verup.sh {version_type}")

    return ver


@task
def compile_requirements(c: Context):
    "Convert requirements.in to requirements.txt and requirements.dev.txt."
    start_time = subprocess.check_output(["date", "+%s"]).decode().strip()
    c.run("uv pip compile requirements.in --output-file=requirements.txt --upgrade")  # --refresh-package
    reqs_time = subprocess.check_output(["date", "+%s"]).decode().strip()
    c.run("uv pip compile requirements.dev.in --output-file=requirements.dev.txt --upgrade")
    end_time = subprocess.check_output(["date", "+%s"]).decode().strip()
    print(f"Req's compilation time: {int(reqs_time) - int(start_time)} seconds")
    print(f"Req's dev compilation time: {int(end_time) - int(reqs_time)} seconds")
    print(f"Total execution time: {int(end_time) - int(start_time)} seconds")

    
    c.run("scripts/include_pyproject_requirements.py requirements.in")
    


@task(pre=[compile_requirements])
def reqs(c: Context):
    """Upgrade requirements including pre-commit."""
    c.run("pre-commit autoupdate")
    c.run("uv pip install -r requirements.dev.txt")

def docs_task_factory(language: str):
    @task
    def docs(c: Context):
        """Docs preview for the language specified."""
        c.run("open -a 'Google Chrome' http://127.0.0.1:8000/aws-s3/")
        c.run(f"scripts/docs-render-config.sh {language}")
        if language != "en":
            shutil.rmtree(f"./docs/src/{language}/images", ignore_errors=True)
            shutil.copytree("./docs/src/en/images", f"./docs/src/{language}/images")
            shutil.copy("./docs/src/en/reference.md", f"./docs/src/{language}/reference.md")
        c.run("mkdocs serve -f docs/_mkdocs.yml")

    return docs


@task
def uv(c: Context):
    """Install or upgrade uv."""
    c.run("curl -LsSf https://astral.sh/uv/install.sh | sh")


@task
def pre(c):
    """Run pre-commit checks"""
    c.run("pre-commit run --verbose --all-files")


namespace = Collection.from_module(sys.modules[__name__])
for name in ALLOWED_VERSION_TYPES:
    namespace.add_task(ver_task_factory(name), name=f"ver-{name}")
for name in ALLOWED_DOC_LANGUAGES:
    namespace.add_task(docs_task_factory(name), name=f"docs-{name}")



import asyncio
import functools
from typing import Iterable, Any, Dict, Optional, Tuple, List

import aiobotocore.session
import aiobotocore.client
from botocore.config import Config


@functools.lru_cache()
def create_session() -> aiobotocore.session.AioSession:
    """Create a session object."""
    return aiobotocore.session.get_session()


def get_s3_client() -> aiobotocore.client.AioBaseClient:
    """Get S3 client."""
    session = create_session()
    config = Config(
        retries={"max_attempts": 3, "mode": "adaptive"},
    )
    return session.create_client("s3", config=config)


class ListObjectsAsync:
    def __init__(self, bucket: str) -> None:
        self._bucket = bucket

    async def _list_objects(
        self,
        s3_client: aiobotocore.client.AioBaseClient,
        prefix: str,
        current_depth: int,
        max_depth: Optional[int],
    ) -> Tuple[Iterable[Dict[str, Any]], Iterable[Tuple[str, int]]]:
        paginator = s3_client.get_paginator("list_objects_v2")
        objects = []
        prefixes = []

        params = {"Bucket": self._bucket, "Prefix": prefix}
        if max_depth is None or current_depth < max_depth:
            params["Delimiter"] = "/"

        async for page in paginator.paginate(**params):
            for obj in page.get("Contents", []):
                key: str = obj["Key"]
                if key.endswith("/"):
                    continue  # Omit "directories"

                objects.append(obj)

            if "Delimiter" in params:
                prefixes.extend(
                    [
                        (prefix["Prefix"], current_depth + 1)
                        for prefix in page.get("CommonPrefixes", [])
                    ]
                )

        return objects, prefixes

    async def list_objects(
        self, prefix: str = "/", max_depth: Optional[int] = None
    ) -> Iterable[Dict[str, Any]]:
        """List all objects in the bucket with given prefix."""
        objects: List[Dict[str, Any]] = []
        tasks = set()

        async with get_s3_client() as s3_client:
            tasks.add(asyncio.create_task(self._list_objects(s3_client, prefix, 0, max_depth)))

            while tasks:
                done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
                tasks = pending

                for task in done:
                    files, folders = await task
                    objects.extend(files)

                    for folder, level in folders:
                        tasks.add(
                            asyncio.create_task(
                                self._list_objects(s3_client, folder, level, max_depth)
                            )
                        )

        return objects



import pytest
from aws_s3 import ListObjectsAsync


@pytest.mark.asyncio
async def test_list_objects(mock_s3_structure):
    walker = ListObjectsAsync("mock-bucket")
    keys = sorted([object["Key"] for object in await walker.list_objects(prefix="root/")])

    expected_keys = sorted([
        'root/data01/image01.png',
        'root/data01/images/img11.jpg',
        'root/data01/docs/doc12.pdf',
        'root/data01/archives/archive13a.zip',
        'root/data01/archives/archive13b.zip',
        'root/data02/report02.docx',
        'root/data02/reports/report21.docx',
        'root/data02/logs/log22.txt',
        'root/data02/scripts/script23.py',
        'root/data03/video03a.mp4',
        'root/data03/video03b.mp4',
        'root/data03/video03c.mp4'
    ])

    assert sorted(keys) == sorted(expected_keys)



from aws_s3 import __version__
from aws_s3.main import aws_s3
from click.testing import CliRunner


def test_version():
    assert __version__


def test_version_option():
    runner = CliRunner()
    result = runner.invoke(aws_s3, ['--version'])
    assert result.exit_code == 0
    assert __version__ in result.output
